#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2012 Liricus SRL <info@liricus.com.ar>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Fra


#===============================================================================
# DOCS
#===============================================================================

"""Application routes that conform the base api.
All the api calls return a JSON object with information on the requested action.

    **JSON:**
        :action:
            The api function called.
        :code:
            HTML status code.
        :error:
            Boolean, ``false`` if nothing weird happened ``true`` otherwise.
        :traceback:
            Apperas when ``error`` is ``true`` and ``DEBUG`` is ``True`` in 
            ``settings`` Python traceback most recent call.
        :msg:
            Appears when ``error`` is ``true``, it's the html error message. 

"""


#===============================================================================
# IMPORTS
#===============================================================================

import string
import random
import datetime

import peewee

import flask

from socialframe_server import core
from socialframe_server import models
from socialframe_server import settings
from socialframe_server import emailutil


#===============================================================================
# ROUTES
#===============================================================================

@core.APP.route("/")
def index():
    """Returns a message if the server work
    
    **return**
        {"error": false, action: "index", "code": 200, "msg": "its works"}
    
    """
    return core.json_data(msg="its works")


@core.APP.route("/sizes")
def sizes():
    """Return al aivailable sizes for images
    
    ***return**
        {"error": false, action: "sizes", "code": 200, "sizes": [...]}
        
    """
    return core.json_data(sizes=settings.IMAGE_SIZES)
    

@core.APP.route("/apihelp")
def apihelp():
    """Returns api documentation and usage.
    
    """
    
    data = {}
    for route in core.APP.url_map.iter_rules():
        if route.rule.startswith("/static/"):
            continue
        doc = getattr(core.APP.view_functions[route.endpoint], "__doc__", None)
        if doc:
            doc = doc.splitlines()[0]
        else:
            doc = "<NO-DOC>"
        rd = {
            "args": list(route.arguments),
            "doc": doc,
        }
        data[core.absurl(route.rule)] = rd
    return core.json_data(**data)


@core.APP.route("/register_user", methods=['POST'])
@core.transactional
def register_user():
    """Register a new user into the server.
        
        **Parameters**  
            :email: User email.
            :fname: First name of the user.
            :lname: Last name
            :birth: Date with format YYYYMMDD (19860808 for 1986-08-08) format.
            :sex: M or F.
            :password: password.
        
        Example call:
            ``/register_user/hello@world.com/foo/bar/19860808/m/mypassword``
        
        **Return**
            
            {"error": false, action: "register_user", "code": 200}
            
            Also an email is send to a user with a link to confirm registration.
        
    """
    
    user = models.User()
    user.email = flask.request.form["email"]
    user.fname = flask.request.form["fname"]
    user.lname = flask.request.form["lname"]
    user.birth = core.parsedate(flask.request.form["birth"]).date()
    user.sex = flask.request.form["sex"].lower()
    user.set_password(flask.request.form["password"])
    user.save()
    profile = models.Profile.get(user=user)
    profile.operation = profile.OPERATION_CREATE_USER
    profile.save()
    
    data = core.absurl("confirm_registration", user.email, profile.operation_hash)
    emailutil.send_mail(settings.MAIL_CREATION, user,
                        flask.request.accept_languages.values(),
                        data)
    
    return core.json_data()


@core.APP.route("/confirm_registration/<email:email>/<userhash>")
def confirm_registration(email, userhash):
    """To confirm registrations. If ``email`` it's not found throws ``Error 403``.
        
        **Parameters**
            :email:
                User email.
            :userhash:
                User hash.
        
        **Return**

            {"action": "confirm_registration", "code": 200, "error": false}
    
    """
    
    user = models.User.get(email=email)
    if not user.active:
        profile = models.Profile.get(user=user)
        if profile.validate_hash(profile.OPERATION_CREATE_USER, userhash):
            user.status = user.STATUS_ACTIVE
            user.last_know_lang = (flask.request.accept_languages.values()[0]
                                   if flask.request.accept_languages.values()
                                   else settings.DEFAULT_LANG)
            user.last_login = datetime.datetime.now()
            user.save()
        else:
            flask.abort(403)
    return core.json_data()


@core.APP.route("/reactivate_user/<email:email>/<userhash>")
def reactivate_user(email, userhash):
    """To reactivate user is is suspended. If ``email`` it's not found throws ``Error 403``.
        
        **Parameters**
            :email:
                User email.
            :userhash:
                User hash.
        
        **Return**

            {"action": "reactivate_user", "code": 200, "error": false}
    
    """
    user = models.User.get(email=email)
    if not user.active:
        profile = models.Profile.get(user=user)
        if profile.validate_hash(models.Profile.OPERATION_REACTIVATE_USER, userhash):
            user.status = user.STATUS_ACTIVE
            user.last_know_lang = (flask.request.accept_languages.values()[0]
                                   if flask.request.accept_languages.values()
                                   else settings.DEFAULT_LANG)
            user.last_login = datetime.datetime.now()
            user.save()
        else:
            flask.abort(403)
    return core.json_data()



@core.APP.route("/change_user", methods=["POST"])
def change_user():
    """Changes the values of the current user
    
    ``email`` are immutable and raise ``error 500``
    
    **Parameters**
            data to be modified
            
        **Return**

            {"action": "change_user", "code": 200, "error": false}
    
    """
    
    user = core.APP_AUTH.get_logged_in_user()
    for k, v in flask.request.form.items():
        if k.startswith("_") or k in models.User.FORBIDEN:
            flask.abort(500)
        if k == "password":
            user.set_password(v)
        else:
            field = getattr(models.User, k)
            if not isinstance(field, peewee.Field):
                flask.abort(500)
            fieldtype = type(field)
            typed = models.FIELD_CONVERSORS[fieldtype](v)
            setattr(user, "__" + k, typed)
        
    user.save()
    return core.json_data()
        
        
@core.APP.route("/change_bubble", methods=["POST"])
def change_bubble():
    """Changes the values of a bubble
        
    **Parameters**
            :bubble_name: string
            and data to be modified
            
        **Return**

            {"action": "change_bubble", "code": 200, "error": false}
    
    """
    data = dict(flask.request.form)
    bubble_name = data.pop("bubble_name")[0]
    user = core.APP_AUTH.get_logged_in_user()
    bubble = models.Bubble.get(name=bubble_name, user=user)
    for k, v in data.items():
        v = v[0]
        if k.startswith("_") or k in models.Bubble.FORBIDEN:
            flask.abort(500)
        if k == "password":
            bubble.set_password(v)
        elif k == "tags":
            bubble.set_tags(v.split(","))
        else:
            field = getattr(models.Bubble, k)
            if not isinstance(field, peewee.Field):
                flask.abort(500)
            fieldtype = type(field)
            typed = models.FIELD_CONVERSORS[fieldtype](v)
            setattr(bubble, "__" + k, typed)
    bubble.save()
    return core.json_data()
    

@core.APP.route("/is_loggedin")
def is_loggedin():
    """Checks for current user to be logged.
        The following variable ``is_loggedin[Boolean]`` is returned in 
        the JSON as well.
        
        
        **Return**
            
            {"action": "is_loggedin","code": 200,"is_loggedin": true,"error": false}
    
    """
    
    loggedin = bool(core.APP_AUTH.get_logged_in_user())
    return core.json_data(is_loggedin=loggedin)


@core.APP.route("/login", methods=["POST"])
@core.transactional
def login():
    """User Login. If ``email`` it's not found throws ``Error 500``
        
        **Parameters**
        
            :email:
                User email.
            :password:
                User email password.
        
        
        **Return**
            
            {"error": false, action: "login", "code": 200}

    """
    email = flask.request.form["email"]
    password = flask.request.form["password"]
    
    user = models.User.get(email=email, active=True)
    if not user or not user.check_password(password):
        flask.abort(403)
    core.APP_AUTH.login_user(user)
    user.last_login = datetime.datetime.now()
    user.last_know_lang = (flask.request.accept_languages.values()[0]
                           if flask.request.accept_languages.values()
                           else settings.DEFAULT_LANG)
    user.save()
    return core.json_data()


@core.APP.route("/logout")
@core.login_required
def logout():
    """Logout user. If no user it's logged throws ``Error 403``.
        
        **return**
        
            {"error": false, action: "logout", "code": 200}

    """
    core.APP_AUTH.logout_user(core.APP_AUTH.get_logged_in_user())
    return core.json_data()


@core.APP.route("/reset_password", methods=["POST"])
@core.transactional
def reset_password():
    """If ``email`` exist resets its password otherwise throws ``Error 403``.
        
        **Parameters**
            
            :email:
                User email.
        
        **Return**
            
            {"action": "reset_password", "code": 200, "error": false}
        
            Also an email is sent with the new password.
    
    """
    
    def gen_password(size=8):
        chars = string.letters + string.digits
        newpasswd = ""
        for i in range(size):
            newpasswd += random.choice(chars)
        return newpasswd
    
    user = models.User.get(email=flask.request.form["email"])
    if not user:
        flask.abort(403)
    new_password = gen_password()
    user.set_password(new_password)
    user.save()
    
    emailutil.send_mail(settings.MAIL_RESET_PASSWORD, user,
                        flask.request.accept_languages.values(),
                        new_password)
                    
    return core.json_data()


@core.APP.route("/delete_user", methods=["POST"])
@core.login_required
@core.transactional
def delete_user():
    """Deletes current user.
        If not logged in trhows ``Error 403``.
        
        **Params**
            :password: the actual logged in user
        
        **Return**
            
            {"action": "delete_user","code": 200,"error": false}
            
            Also an email is sent to confirm the action.
    
    """

    user = core.APP_AUTH.get_logged_in_user()
    if not user.check_password(flask.request.form["password"]):
        flask.abort(403)
    profile = models.Profile.get(user=user)
    profile.operation = models.Profile.OPERATION_DELETE_USER
    profile.save()

    data = core.absurl("confirm_delete", user.email, profile.operation_hash)
    emailutil.send_mail(settings.MAIL_DELETE, user,
                        flask.request.accept_languages.values(),
                        data)
    
    return core.json_data()


@core.APP.route("/confirm_delete/<email:email>/<deletehash>")
@core.login_required
@core.transactional
def confirm_delete(email, deletehash):
    """Confirms user ``email`` and goes ahead with deletion.
        The ``Error 403`` is shown when:
            * ``email`` to delete difers to the user ``email``.
            * ``deletehash`` it's not equal to the operation.
    
        **Parameters**
        
            :email:
                User email.
            :deletehash:
                Operation hash.
        
        **Return**
            
            {"action": "confirm_delete","code": 200,"error": false}
    
    """
    
    user = core.APP_AUTH.get_logged_in_user()
    if user.email != email:
        flask.abort(403)
    profile = models.Profile.get(user=user)
    if profile.validate_hash(profile.OPERATION_DELETE_USER, deletehash):
        user.delete_instance(recursive=True)
        profile.delete_instance(recursive=True)
    else:
        flask.abort(403)
    return core.json_data()


@core.APP.route("/create_bubble", methods=["POST"])
@core.login_required
@core.transactional
def create_bubble():
    """Creates a new bubble to the current user.
        
        **Parameters**
        
            :name:
                Name of the bubble.
            :listed[true, false]:
                Is bubble listed or not?.
            :desc[string, null]:
                A description of the bubble.
            :tags[list, null]:
                Associated tags.
            :password[string, null]:
                Bubble password.
            
        **Return**
        
            {"action": "create_bubble", "code": 200, "error": false}
    
    """
    
    user = core.APP_AUTH.get_logged_in_user()
    bubble = models.Bubble()
    bubble.user = user
    bubble.name = flask.request.form["name"]
    bubble.desc = flask.request.form["desc"]
    if flask.request.form["listed"] == "true":
        bubble.listed = True  
    if flask.request.form["password"] != "null":
        bubble.set_password(flask.request.form["password"])
    bubble.save()
    if flask.request.form["tags"] != "null":
        tags = flask.request.form["tags"].split(",")
        bubble.set_tags(tags)
    return core.json_data()


@core.APP.route("/list_bubbles/<email:email>")
def list_bubbles(email):
    """Returns only listed ``bubbles`` from  user ``email``.
        
        **Parameters**
            :email:
                User email.
        
        **Return**
            
            {
              "action": "list_bubbles",
              "bubbles": {
                "test1": {
                  "has_password": true,
                  "listed": true,
                  "tags": []
                },
                "test3": {
                  "has_password": false,
                  "listed": true,
                  "tags": []
                }
              },
              "code": 200,
              "error": false
            }
    
    """
    
    data = {}
    for bubble in models.Bubble.filter(user__email=email, listed=True):
        data[bubble.name] = {
            "has_password": bubble.has_password,
            "listed": bubble.listed,
            "tags": bubble.tags,
        }
    return core.json_data(bubbles=data)


@core.APP.route("/my_bubbles")
@core.login_required
def my_bubbles():
    """Returns all the current user bubbles.
    
        **Return**
            {
              "action": "my_bubbles", 
              "code": 200, 
              "my_bubbles": {
                "test1": {
                  "has_password": true, 
                  "listed": true, 
                  "tags": []
                }, 
                "test3": {
                  "has_password": false, 
                  "listed": true, 
                  "tags": []
                }, 
                "test2": {
                  "has_password": true, 
                  "listed": false, 
                  "tags": []
                }
              }, 
              "error": false
            }
            
    """
    
    user = core.APP_AUTH.get_logged_in_user()
    data = {}
    for bubble in models.Bubble.filter(user=user):
        data[bubble.name] = {
            "has_password": bubble.has_password,
            "listed": bubble.listed,
            "tags": bubble.tags,
        }
    return core.json_data(bubbles=data)


@core.APP.route("/validate_bubble_password/<email:email>/<bubble_name>/<null:password>")
def validate_bubble_password(email, bubble_name, password):
    user = models.User.get(email=email)
    bubble = models.Bubble.get(user=user, name=bubble_name)
    bpass = bubble.check_password(password)
    return core.json_data(valid_password=bpass)
    

@core.APP.route("/list_photos/<email:email>/<bubble_name>/<null:password>")
def list_photos(email, bubble_name, password):
    """Returns a list of the user (``email``) photos and location in the specified ``bubble_name``.
        If the bubble has a password and the one given does not match, ``Error 403`` is raised.
        
        **Parameters**
        
            :email:
                User email.
            :bubble_name:
                Bubble name.
            :password[string, null]:
                Password.
        
        **Return**
            A JSON listing the photos in the bubble:
                {
                  "action": "list_photos", 
                  "photos": {
                    "a7e804d33aa585bmodels.30bec77896fa9512eea797.jpg": {
                      "sha": "f45cf037f11ca3ade43f59be3bb0c410b5b0fbf5"
                    }, 
                    "1567103d4b9cf570e8b38b281725c9049ef24f64.jpg": {
                      "sha": "0ca6bbd29dda257939700559d85a593939ce5efb"
                    }, 
                    "ff77f365484models.647b7df9fdade019a3be9de11e.jpg": {
                      "sha": "37b8458f39be3ea3a1a8b221895d4a586eca1663"
                    }, 
                    "0b2a5242381f1942c838322a8000e65c8680aada.jpg": {
                      "sha": "f0634007fe98125b50d4718300a395dafbfb6baa"
                    }
                  }, 
                  "error": false
                }
    
    """
    
    photos = {}
    user = models.User.get(email=email)
    bubble = models.Bubble.get(user=user, name=bubble_name)
    
    if bubble.has_password and not bubble.check_password(password):
        flask.abort(403)
    for image in models.Image.filter(bubble=bubble):
        photos[image.name] = {
            "sha": image.sha
        }
    return flask.jsonify(error=False, action="list_photos", photos=photos)


@core.APP.route("/delete_bubble/<bubble_name>")
@core.login_required
@core.transactional
def delete_bubble(bubble_name):
    """Deletes a bubble from the current user.
        
        **Parameters**
            :bubble_name:
                Name of bubble.
        
        **Return**
            
            {"action": "delete_bubble", "code": 200, "error": false}
            
    """
    
    user = core.APP_AUTH.get_logged_in_user()
    bubble = models.Bubble.get(name=bubble_name, user=user)
    bubble.delete_instance(recursive=True)
    return core.json_data()


@core.APP.route("/get_photo/<email:email>/<bubble_name>/<filename>/<size>/<null:password>")
@core.transactional
def get_photo(email, bubble_name, filename, size, password):
    """Looks for a photo (``filename``) in the specified ``bubble_name`` of the user ``email`` and returns the photo.
        If the bubble has a password and the one given does not match, ``Error 403`` is raised.
        
        **Parameters**
            :email:
                User email.
            :bubble_name:
                Bubble name.
            :filename:
                Photo name.
            :password[string, null]:
                Password.
    
    """
    if size not in settings.IMAGE_SIZES:
        flask.abort(400)
    
    user = models.User.get(email=email)
    bubble = models.Bubble.get(user=user, name=bubble_name)
    
    if bubble.has_password and not bubble.check_password(password):
        flask.abort(403)

    image = models.Image.get(bubble=bubble, name=filename)
    return flask.send_file(image.get_size(size).path, as_attachment=True)
    

@core.APP.route("/del_photo/<bubble_name>/<filename>")
@core.login_required
@core.transactional
def del_photo(bubble_name, filename):
    """Deletes a photo from a bubble
        
        **Parameters**
            :bubble_name:
                Name of bubble.
            :filename:
                Photo name.
    
    """
    
    user = core.APP_AUTH.get_logged_in_user()
    bubble = models.Bubble.get(user=user, name=bubble_name)
    image = models.Image.get(bubble=bubble, name=filename)
    image.delete_instance(True)
    return core.json_data()
    
    
@core.APP.route("/upload_photo/<bubble_name>", methods=["POST"])
@core.login_required
@core.transactional
def upload_photo(bubble_name):
    """Uploads a photo to the specified bubble.
        
        **Parameters**
            :bubble_name:
                Name of the bubble to upload
    
    """
    
    bubble = models.Bubble.get(name=bubble_name)
    
    storage = flask.request.files['photo']
    
    img = models.Image()
    img.bubble = bubble
    img.save()
    
    img.store(storage)
    img.save()
    
    return core.json_data()


@core.APP.route("/give_me_bubbles")
def give_me_bubbles():
    """Returns random listed bubbles.
        
        **Return**
            {
              "action": "give_me_bubbles", 
              "bubbles": {
                "superburbu": {
                  "email": "hello@world.com", 
                  "has_password": false, 
                  "listed": true, 
                  "tags": []
                }, 
                "test3": {
                  "email": "foo@bar.com", 
                  "has_password": false, 
                  "listed": true, 
                  "tags": []
                }, 
                "test5": {
                  "email": "foo@bar.com", 
                  "has_password": false, 
                  "listed": true, 
                  "tags": []
                }, 
                "test6": {
                  "email": "foo@bar.com", 
                  "has_password": false, 
                  "listed": true, 
                  "tags": []
                }
              }, 
              "code": 200, 
              "error": false
            }
    
    """
    
    data = {}
    for bubble in models.Bubble.random_bubbles():
        data[bubble.name] = {
            "has_password": bubble.has_password,
            "listed": bubble.listed,
            "tags": bubble.tags,
            "email": bubble.user.email,
        }
    return core.json_data(bubbles=data)


@core.APP.route("/find/<what>/<list:arguments>")
def find(what, arguments):
    """It finds users or bubbles that match the arguments given.
        
        **Parameters**
            :what:
                ``users`` to look for users. If no match is found the ``users``
                field is returned empty.
                ``bubbles`` to look for bubbles.
                Other keyword raise ``Error 404``
            :argumens[list]:
                Keywords to look for. Several arguments separated by ','.
        
        **Return**
        
            for example: the next JSON corresponds with the call /users/Foobar
                {
                  "action": "find",
                  "code": 200,
                  "users": {
                    "hello@world.com": {
                      "lname": "python",
                      "birth": "2012/08/07",
                      "fname": "Foobar",
                      "sex": "m"
                    }
                  },
                  "error": false
                }

    """
    
    found = {}
    for arg in arguments:
        if what == "users":
            query = models.User.select()
            query = query.where(peewee.Q(email__icontains=arg) 
                                | peewee.Q(fname__icontains=arg) 
                                | peewee.Q(lname__icontains=arg))
            for u in query:
                found[u.email] = {"fname": u.fname,
                                  "lname":u.lname,
                                  "sex":u.sex,
                                  "birth": u.birth.strftime(settings.DATE_FORMAT)}
        elif what == "bubbles":
            query = models.Bubble.select()
            query = query.where(peewee.Q(name__icontains=arg)
                                | peewee.Q(desc__icontains=arg)
                                | peewee.Q(listed=True))
            for b in query:
                found[b.name] = {"has_password": b.has_password,
                                 "tags": b.tags,
                                 "email": b.user.email}
        else:
            flask.abort(404)
    data = {what: found}
    return core.json_data(**data)
            
                

#===============================================================================
# MAIN
#===============================================================================

if __name__ == "__main__":
    print(__doc__)
